/**
 * @file Main Module
 * @author Patryk Gliszczynski
 * @version 1.0
 */

class App {
  /**
   * Main module responsible for the maintanance of the core animation components.
   */

  constructor() {
    this.scene = this.Scene();
    this.camera = this.Camera();
    this.controls = this.Controls();
    this.webGLRenderer = this.WebGLRenderer();
    this.css3DRenderer = this.CSS3DRenderer();

    this.simulation = new Simulation(new World(this.camera, this.scene));

    document.body.appendChild(this.webGLRenderer.domElement);
    document.body.appendChild(this.css3DRenderer.domElement);
    window.addEventListener("resize", this.onWindowResize.bind(this), false);
    return this;
  }

  Scene() {
    /**
     * Creates a new Scene object.
     */
    let scene = new THREE.Scene();
    scene.fog = new THREE.Fog(Config.Fog.BACKGROUND, Config.Fog.NEAR, Config.Fog.FAR);
    return scene;
  }

  Camera() {
    /**
     * Creates a new Camera object.
     */
    let camera = new THREE.PerspectiveCamera(Config.Camera.FOV, Config.Camera.ASPECT, Config.Camera.NEAR, Config.Camera.FAR);
    camera.position.set(Config.Camera.X_POS, Config.Camera.Y_POS, Config.Camera.Z_POS);
    camera.updateProjectionMatrix();
    return camera;
  }

  Controls() {
    /**
     * Creates a new Control object alowing to steer the camera.
     */
    let controls = new THREE.OrbitControls(this.camera);
    controls.autoRotate = Config.Controls.AUTO_ROTATE;
    controls.autoRotateSpeed = Config.Controls.AUTO_ROTATE_SPEED;
    controls.minDistance = Config.Controls.MIN_DISTANCE;
    controls.maxDistance = Config.Controls.MAX_DISTANCE;
    controls.maxPolarAngle = Config.Controls.MAX_POLAR_ANGLE;
    controls.update();
    return controls
  }

  WebGLRenderer() {
    /**
     * Creates a WebGL Renderer object used to visualize 3D components.
     */
    let renderer = new THREE.WebGLRenderer({antialias: Config.WebGLRenderer.ANTIALIAS});
    renderer.setSize(Config.WebGLRenderer.WIDTH, Config.WebGLRenderer.HEIGHT);
    renderer.setPixelRatio(Config.WebGLRenderer.PIXEL_RATIO);
    renderer.shadowMap.enabled = Config.WebGLRenderer.ShadowMap.ENABLED;
    renderer.shadowMap.type = THREE.PCFSoftShadowMap;
    renderer.shadowMap.renderSingleSided = Config.WebGLRenderer.ShadowMap.SINGLE_SIDED;
    renderer.gammaInput = Config.WebGLRenderer.GAMMA_INPUT;
    renderer.gammaOutput = Config.WebGLRenderer.GAMMA_OUTPUT;
    return renderer;
  }

  CSS3DRenderer() {
    /**
     * Creates a CSS3D Renderer object used to visualize HTML+CSS components in 3D scene.
     */
    let renderer = new THREE.CSS3DRenderer();
    renderer.setSize(Config.CSS3DRenderer.WIDTH, Config.CSS3DRenderer.HEIGHT);
    renderer.domElement.style.position = Config.CSS3DRenderer.POSITION;
    renderer.domElement.style.top = Config.CSS3DRenderer.TOP;
    return renderer
  }

  run() {
    /**
     * Starts the simulation and rendering process.
     */
    this.render();
    this.simulation.run();
  }

  render() {
    /**
     * Main renderer loop, called on each frame.
     */
    requestAnimationFrame(this.render.bind(this));
    TWEEN.update();
    this.controls.update();
    this.webGLRenderer.render(this.scene, this.camera);
    this.css3DRenderer.render(this.scene, this.camera);
  }

  onWindowResize() {
    /**
     * Updates the dimmensions of renderers after window size changes.
     */
    this.camera.aspect = window.innerWidth / window.innerHeight;
    this.camera.updateProjectionMatrix();
    this.webGLRenderer.setSize(window.innerWidth, window.innerHeight);
    this.css3DRenderer.setSize(window.innerWidth, window.innerHeight);
  }
}

new App().run();
